from typing import Callable, Optional

from .._autoflow.manager import Flow
from ..constants import DEFAULT_CORE_HOST
from ..interpreter.core import CoreInterpreter
from ..models.platform import Platform
from ..models.task import Task


def flow(
    interpreter: Optional[Platform] = None,
    core_host: Optional[str] = DEFAULT_CORE_HOST,
    core_auth: Optional[tuple[str, str]] = None,
) -> Callable[[], tuple[Task, object]]:
    def wrapper(function: Callable[[], None]) -> Callable[[], str]:
        def fn(*args, **kwargs):  # noqa: ANN202, ANN002, ANN003
            match (interpreter):
                case Platform.CORE:
                    __tree = None
                    with Flow() as tree:
                        function(*args, **kwargs)
                        __tree = tree
                    __interpreter = CoreInterpreter(
                        core_host=core_host, core_auth=core_auth
                    )
                    __task, __cfg = __interpreter.interpret(__tree)
                    return Task(
                        alias=function.__name__,
                        task_id=__task,
                        cfg_id=__cfg,
                        host=core_host
                    ), __interpreter.state
                case _:
                    raise NotImplementedError(
                        f"Interpreter {interpreter} is not implemented"
                    )
        return fn
    return wrapper



from typing import Optional

from pydantic import BaseModel

from malevich.constants import DEFAULT_CORE_HOST
from malevich.models.manifest import Dependency


class ImageOptions(BaseModel):
    image_ref: str
    image_auth_user: Optional[str]
    image_auth_pass: Optional[str]
    core_host: Optional[str] = DEFAULT_CORE_HOST
    core_auth_user: Optional[str]
    core_auth_token: Optional[str]
    checksum: str
    operations: dict[str, str] = None


class ImageDependency(Dependency):
    options: ImageOptions



from malevich.models.platform import Platform

DEFAULT_CORE_HOST = "https://core.malevich.ai"

APP_HELP = """
Welcome to [bold magenta]Malevich Metascript[/bold magenta]! This is a CLI tool that
allows you to create and run multitask pipelines using [yellow]Malevich API[/yellow].
\n _____ \n
The tool includes[i] autoflow [/i]engine, that analyzes your code and turns it into a human-readable
pipeline;[i] installers [/i]that enable you to populate your flow with a huge variety of pre-built
components; and finally[i] interpreters [/i]that run your code on Malevich backend freeing you
from installing any dependency.

How you start:
\t1. Run [code] malevich use [/code] to install apps and get access to their functionality
\t2. Write a pipeline using simple Python scripts and [code] @flow [/code] decorators
\t3. Connect to Malevich API and run your flow effortlessly

Check out [link=https://docs.malevich.ai/malevich]Malevich documentation[/link] for more details
"""  # noqa: E501

USE_HELP = """
Install apps to get access to the processors they provide
"""

USE_IMAGE_HELP = """
Install apps using published Docker images. The information about
the functionality provided by them will be pulled and parsed accordingly.
"""

IMAGE_BASE = "public.ecr.aws/u6e1k0c7/{app}:latest"


CorePlatform = Platform.CORE
SpacePlatofrm = Platform.SPACE



import rich
import typer
from rich.progress import Progress, SpinnerColumn, TextColumn

from .commands.manifest import app as manifest_app
from .commands.use import install_from_image
from .commands.use import use as use_app
from .constants import APP_HELP
from .install.image import ImageInstaller
from .manifest import ManifestManager

app = typer.Typer(help=APP_HELP, rich_markup_mode="rich")

@app.command('install')
def auto_use(package_name: str) -> None:
    rich.print(f"\n\nAttempting automatic installation of package [b blue]{package_name}[/b blue]\n")  # noqa: E501
    with Progress(
        SpinnerColumn(),
        TextColumn("{task.description}"),
    ) as progress:
        task = progress.add_task(
            f"Attempting to install [b blue]{package_name}[/b blue] with [i yellow]image[/i yellow] installer", total=1  # noqa: E501
        )
        try:
            install_from_image(package_name=package_name)
        except Exception:
            # BUG: This is not a good way to handle errors
            progress.update(
                task,
                description="[red]✘[/red] Failed with [yellow]image[/yellow] installer",
                completed=1
            )
            progress.stop()
            rich.print("\n\n[red]Installation failled[/red]")
        else:
            progress.update(
                task,
                description="[green]✔[/green] Success with [yellow]image[/yellow]",
                completed=1
            )
            progress.stop()
            rich.print(f"\n\nInstallation of [blue]{package_name}[/blue] with [yellow]image[/yellow] installer was [green]successful[/green]")  # noqa: E501
        return


@app.command(name="restore")
def restore() -> None:
    manf = ManifestManager()
    image_installer = ImageInstaller()
    with Progress(SpinnerColumn(), TextColumn("{task.description}")) as progress:
        for record in manf.query("dependencies"):
            key = [*record.keys()][0]
            dependency = record[key]
            task = progress.add_task(
                f"Package [green]{dependency['package_id']}[/green]", total=1
            )
            installed_by = dependency["installer"]
            if installed_by == "image":
                parsed = image_installer.construct_dependency(dependency)
                parsed.options.image_auth_pass = manf.query_secret(
                    parsed.options.image_auth_pass,
                    only_value=True,
                )
                parsed.options.image_auth_user = manf.query_secret(
                    parsed.options.image_auth_user,
                    only_value=True,
                )
                parsed.options.core_auth_user = manf.query_secret(
                    parsed.options.core_auth_user,
                    only_value=True,
                )
                parsed.options.core_auth_token = manf.query_secret(
                    parsed.options.core_auth_token,
                    only_value=True,
                )

                # if not all(
                #     [
                #         parsed.options.image_auth_pass,
                #         parsed.options.image_auth_user,
                #         parsed.options.core_auth_user,
                #         parsed.options.core_auth_token,
                #     ]
                # ):
                #     rich.print(
                #         "\n[red bold]Error[/red bold]: \n"
                #         "\tIt seems you are missing a number of secrets. Please "
                #         "fix it using command [code]malevich manifest secrets restore[/code]\n\n"  # noqa: E501
                #     )
                #     quit(-1)

                # if not parsed.options.image_ref:
                #     rich.print(
                #         "\n[red bold]Error[/red bold]: \n"
                #         "\tMissing [code]image_ref[/code] variable. Please"
                #         "\tfix it manually "  # noqa: E501
                #     )
                image_installer.restore(parsed)
            progress.advance(task, 1)

app.add_typer(use_app, name="use")
app.add_typer(manifest_app, name="manifest")

def main() -> None:
    app()


if __name__ == "__main__":
    main()



from abc import ABC, abstractmethod


class Anomaly(ABC):
    @abstractmethod
    def __str__(self) -> str:
        pass

class MultipleLeaves(Anomaly):
    def __str__(self) -> str:
        return "multiple leaves are not supported"


class DetectedAnomalyError(Exception):
    """Exception raised when an anomaly is detected"""

    def __init__(self, anomaly: Anomaly) -> None:
        self.anomaly = anomaly
        super().__init__(f"During the execution of the workflow, an anomaly was detected: {anomaly}")  # noqa: E501



from typing import Optional
from uuid import uuid4

import pandas as pd

from .._autoflow import tracer as gn  # engine
from ..models.collection import Collection


def collection(
    name: Optional[str] = None,
    file: Optional[str] = None,
    data: Optional[pd.DataFrame] = None
) -> Collection:
    """Creates a collection from a file or a dataframe

    Args:
        name (Optional[str], optional): Name of the collection. Defaults to None.
        file (Optional[str], optional): Path to the file. Defaults to None.
        data (Optional[pd.DataFrame], optional): Dataframe. Defaults to None.

    Raises:
        AssertionError: If both file and data are provided

    Returns:
        Collection: an object that is subject to be passed to processors
    """
    assert any([file is not None, data is not None]), \
        "Either file or data must be provided"

    if file:
        assert data is None, "Cannot provide both file and data"
        data = pd.read_csv(file)

    name = name or uuid4().hex

    return gn.tracer((name, {
            '__collection__': Collection(collection_id=name, collection_data=data),
        }
    ))



import warnings
from typing import Callable, Generic, Optional, TypeVar, Union

from jls_utils import AppLogs, task_prepare, task_run, task_stop
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn

from malevich.constants import DEFAULT_CORE_HOST

T = TypeVar("T")
Callback = TypeVar("Callback", bound=Callable[["Task"], None])


class Task(Generic[T]):
    def __init__(
        self,
        alias: str,
        task_id: str,
        cfg_id: str,
        payload: Optional[T] = None,
        host: Optional[tuple[str, str]] = DEFAULT_CORE_HOST,
    ) -> None:
        self.__task_id = task_id
        self.__cfg = cfg_id
        self.__host = host
        self.__operation_id = None
        self.__payload = payload
        self.__prepared = False
        self.__finished = False
        self.__alias = alias

        self.__callbacks: dict[str, list[Callback]] = {
            "onerr": [lambda t: t.stop()],
        }
        self.__progress = Progress(
            SpinnerColumn(), TextColumn("{task.description} | "), TimeElapsedColumn()
        )

    def onerror(self, callback: Callback) -> None:
        self.__callbacks["onerr"].append(callback)

    def prepare(
        self,
        no_callbacks: bool = False,
        quiet: bool = False,
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> tuple[bool, Union[AppLogs, Exception]]:
        """Prepares the task to be executed at Malevich API.

        Suppresses all errors and returns an Exception in case of failure.

        Args:
            no_callbacks (bool, optional):
                If True, callbacks will not be called on error. Defaults to False.
            quiet (bool, optional):
                If True, no logs will be printed. Defaults to False.
            *args:
                Positional arguments to be passed to the `task_run` function of
                Malevich API.
            **kwargs:
                Keyword arguments to be passed to the `task_run` function of
                Malevich API.

        Returns:
            tuple[bool, Union[AppLogs, Exception]]:
                A tuple of success status and either AppLogs
                (in case of success) or Exception (in case of failure)
        """
        try:
            # Starting task to display progress in
            # CLI interface
            if not quiet:
                __t = self.__progress.add_task(
                    f"Preparing operation [i b yellow]{self.__alias}[/i b yellow]",
                    total=1
                )
                self.__progress.start()

            if 'conn_url' in kwargs:
                warnings.warn(
                    "Keyword argument `conn_url` for `task_prepare` is overriden "
                    "by `host` argument of `Task` class.")
                kwargs.pop('conn_url')

            # Preparing task, if no errors, updating
            # progress bar
            __out = (
                True,
                task_prepare(
                    self.__task_id,
                    cfg_id=self.__cfg,
                    conn_url=self.__host,
                    *args,
                    **kwargs
                ),
            )  # saving output to return it later

            self.__operation_id = __out[1].operationId
            if not quiet:

                self.__progress.update(
                    __t, completed=1,
                    refresh=True,
                    description=(
                        f"[green]✔[/green] Preparing task [i b green]{self.__alias}[/i b green] done. "  # noqa: E501
                        f"Operation ID: [i bright_black]{self.__operation_id}[/ i bright_black]"  # noqa: E501
                    )
                )

            self.__prepared = True
            return __out

        except Exception as e:
            # If error occured, calling callbacks
            if not no_callbacks:
                for func in self.__callbacks['onerr']:
                    func(self)
            # Stopping progress bar to prevent
            # further updates
            try:
                self.__progress.update(__t, description="[red]✘[/red] Error occured. Stopping task preparation.")  # noqa: E501
            except Exception:
                pass
            self.__progress.stop()
            return False, e

    def run(
        self,
        no_callbacks: bool = False,
        quiet: bool = False,
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> tuple[bool, Union[AppLogs, Exception]]:
        """Runs the task at Malevich API.

        Suppresses all errors and returns an Exception in case of failure.

        Args:
            no_callbacks (bool, optional):
                If True, callbacks will not be called on error. Defaults to False.
            quiet (bool, optional):
                If True, no logs will be printed. Defaults to False.
            *args:
                Positional arguments to be passed to the `task_run` function of
                Malevich API.
            **kwargs:
                Keyword arguments to be passed to the `task_run` function of
                Malevich API.

        Returns:
            tuple[bool, Union[AppLogs, Exception]]:
                A tuple of success status and either AppLogs
                (in case of success) or Exception (in case of failure)
        """
        try:
            # Starting task to display progress in
            # CLI interface
            if not quiet:
                __t = self.__progress.add_task(
                    f"Running operation [i bright_black]{self.__operation_id}"
                    "[/i bright_black]",
                    total=1
                )
                self.__progress.start()

            # Preparing task, if no errors, updating
            # progress bar
            __out = (
                True,
                task_run(
                    self.__operation_id,
                    cfg_id=self.__cfg,
                    *args,
                    **kwargs
                ),
            )  # saving output to return it later

            if not quiet:
                self.__progress.update(
                    __t, completed=1, refresh=True,
                    description=f"[green]✔[/green] Task run finished. Operation ID: [i bright_black]{self.__operation_id}[/i bright_black]"  # noqa: E501
                )
            self.__prepared = True
            return __out

        except Exception as e:
            # If error occured, calling callbacks
            if not no_callbacks:
                for func in self.__callbacks['onerr']:
                    func(self)
            # Stopping progress bar to prevent
            # further updates
            try:
                self.__progress.update(__t, description="[red]✘[/red] Error occured. Stopping task running.")  # noqa: E501
            except Exception:
                pass
            self.__progress.stop()
            return False, e


    def stop(
        self, *args, **kwargs  # noqa: ANN003, ANN002
    ) -> tuple[bool, Union[AppLogs, Exception]]:
        """Stops the task at Malevich API.

        Args:
            *args:
                Positional arguments to be passed to the `task_run` function of
                Malevich API.
            **kwargs:
                Keyword arguments to be passed to the `task_run` function of
                Malevich API.

        Returns:
            tuple[bool, Union[AppLogs, Exception]]:
                A tuple of success status and either AppLogs
                (in case of success) or Exception (in case of failure)
        """
        if 'conn_url' in kwargs:
            warnings.warn(
                "Keyword argument `conn_url` for `task_stop` is overriden "
                "by `host` argument of `Task` class.")
            kwargs.pop('conn_url')

        try:
            __out = True, task_stop(
                self.__operation_id,
                conn_url=self.__host,
                *args,
                **kwargs
            )

            self.__progress.stop()
            return __out
        except Exception as e:
            return False, e


    @property
    def finished(self) -> bool:
        return self.__finished

    @property
    def prepared(self) -> bool:
        return self.__prepared

    @property
    def task_id(self) -> str:
        return "" + self.__task_id

    @property
    def payload(self) -> str:
        return self.__payload



from typing import Iterable

from .tree import ExecutionTree


class Flow:
    """Orchestrate the execution of the pipeline"""

    # Internal representation of the execution tree
    __flow = None

    def isinflow() -> bool:
        """Check if the current execution is inside a flow"""
        return Flow.__flow is not None

    def flow_ref() -> ExecutionTree:
        """Returns the reference to the current flow in the context"""
        return Flow.__flow


    def __enter__(self) -> ExecutionTree:
        assert Flow.__flow is None, "Flow already exists"
        Flow.__flow = ExecutionTree()
        return Flow.__flow


    def __exit__(self, *args: Iterable) -> None:
        Flow.__flow = None




from collections import deque
from typing import Any, Generic, Iterable, Iterator, TypeVar

T = TypeVar("T")

class ExecutionTree(Generic[T]):
    tree: list[tuple[T, T]] = []


    def put_edge(self, caller: T, callee: T, link: Any = None) -> None:  # noqa: ANN401
        self.tree.append((caller, callee, link))

    def traverse(self) -> Iterator[tuple[T, T, Any]]:
        """Traverse the execution tree in a determenistic order

        Returns:
            Generator[T]: Generator of nodes
        """
        graph = self.tree
        # Mark visited nodes
        visited = [False] * len(graph)

        # Find roots
        roots = [
            (i, x) for i, x in enumerate(graph)
            if not any(y[1] == x[0] for y in graph)
        ]

        # Traverse
        q = deque(maxlen=len(graph))
        for i, r in roots:
            # BFS
            q.append((i, r,))

        while q:
            j, node = q.pop()
            yield node
            visited[j] = True

            q.extend(
                filter(
                    lambda x: x[1][0] == node[1] and not visited[x[0]],
                    enumerate(graph)
                )
            )


    def leaves(self) -> Iterable[T]:
        return (
            x[1] for x in self.traverse()
            if not any(y[0] == x[1] for y in self.tree)
        )



import uuid
from hashlib import sha256
from typing import Any, Generic, Iterable, TypeVar, Union

from pydantic import BaseModel, Field

from .manager import Flow
from .tree import ExecutionTree


class root(BaseModel):   # noqa: N801
    """A root element of the execution tree

    Used as a default value for the owner of the tracer
    """

    id: str = Field(..., default_factory=lambda: uuid.uuid4().hex)

    def __eq__(self, other: "root") -> bool:
        if not isinstance(other, root):
            return False
        return self.id == other.id


    def __hash__(self) -> int:
        return hash(sha256(self.id.encode()).hexdigest)


T = TypeVar("T", bound=Any)
class autoflow(Generic[T]):  # noqa: N801
    """Autoflow is a bridge between the tracer and the execution tree"""
    def __init__(self, tree: ExecutionTree[T]) -> None:
        self._tree_ref = tree
        self._component_ref = None

    def attach(self, component: T) -> None:
        self._component_ref = component

    def calledby(self, caller: 'tracer', argument: str = None) -> None:
        self._tree_ref.put_edge(self._component_ref, caller, argument)


class tracer(Generic[T]):  # noqa: N801
    """Tracer is a special object that is used to track the execution of the pipeline"""
    def __init__(
        self,
        owner: T = root()
    ) -> None:
        assert (
            Flow.isinflow()
        ), "Tried to create a tracked variable outside of a flow"
        self._owner = owner
        self._autoflow = autoflow(Flow.flow_ref())
        self._autoflow.attach(self)

    @property
    def owner(self) -> T:
        return self._owner



class multitracer(tracer, list[T], Generic[T]):   # noqa: N801
    """Multitracer is a tracer for collections"""
    def __init__(
        self,
        owner: T = root(),
        __iterable: Iterable[T] = []
    ) -> None:
        super().__init__(__iterable)
        super().__init__(owner)


def trace(x: T) -> Union[tracer[T], T]:
    """Create a hidden tracer for the given object"""
    if isinstance(x, list):
        return multitracer.__init__(x)
    else:
        tracer.__init__(x)
    return x



import ast
import os

input_functions = ["input_doc", "input_df", "input_true"]


class SearchDecoratorsVisitor(ast.NodeVisitor):
    decorated_functions = []


class ProcessorSearchVisitor(SearchDecoratorsVisitor):
    """Find all calls to `processor()` function decorator."""

    def __init__(self) -> None:
        """Initialize the visitor."""
        self.decorated_functions = []

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None: # noqa: N802
        """Visit all function definitions."""
        # Iterate over all decorators of the function
        for decorator in node.decorator_list:
            # Check if the decorator is a call
            if isinstance(decorator, ast.Call):
                if (
                    # HACK: Assume that the decorator is a call to `processor` function
                    isinstance(decorator.func, ast.Attribute)
                    and decorator.func.attr == "processor"
                ):
                    self.decorated_functions.append(node.name)
        self.generic_visit(node)


class InputSearchVisitor(SearchDecoratorsVisitor):
    """Find all calls to `input_*()` function decorators."""

    def __init__(self) -> None:
        self.decorated_functions = []

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None: # noqa: N802
        # Iterate over all decorators of the function
        for decorator in node.decorator_list:
            # Check if the decorator is a call
            if isinstance(decorator, ast.Call):
                if (
                    # HACK: Assume that the decorator is a call to `input_*` function
                    isinstance(decorator.func, ast.Attribute)
                    and decorator.func.attr in input_functions
                ):
                    self.decorated_functions.append(node.name)
        self.generic_visit(node)


class InitSearchVisitor(SearchDecoratorsVisitor):
    """Find all calls to `init()` function decorator."""

    def __init__(self) -> None:
        self.decorated_functions = []

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None: # noqa: N802
        # Iterate over all decorators of the function
        for decorator in node.decorator_list:
            # Check if the decorator is a call
            if isinstance(decorator, ast.Call):
                if (
                    # HACK: Assume that the decorator is a call to `init` function
                    isinstance(decorator.func, ast.Attribute)
                    and decorator.func.attr == "init"
                ):
                    self.decorated_functions.append(node.name)
        self.generic_visit(node)


class OutputSearchVisitor(SearchDecoratorsVisitor):
    """Find all calls to `output()` function decorator."""

    def __init__(self) -> None:
        self.decorated_functions = []

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None: # noqa: N802
        # Iterate over all decorators of the function
        for decorator in node.decorator_list:
            # Check if the decorator is a call
            if isinstance(decorator, ast.Call):
                if (
                    # HACK: Assume that the decorator is a call to `output` function
                    isinstance(decorator.func, ast.Attribute)
                    and decorator.func.attr == "output"
                ):
                    self.decorated_functions.append(node.name)
        self.generic_visit(node)


def _scan_tree(
    file: os.PathLike,
    visitor: SearchDecoratorsVisitor,
    skip_syntax_errors: bool = True,
) -> list[str]:
    with open(file) as f:
        try:
            tree = ast.parse(f.read())
        except SyntaxError:
            if skip_syntax_errors:
                return []
            else:
                raise

    visitor.visit(tree)
    return visitor.decorated_functions


def scan_processors(file: os.PathLike) -> list[str]:
    # Traverse the python file and find
    # all calls to `processor()`
    # function decorator

    return _scan_tree(file, ProcessorSearchVisitor())


def scan_inputs(file: os.PathLike) -> list[str]:
    # Traverse the python file and find
    # all calls to `processor()`
    # function decorator

    return _scan_tree(file, InputSearchVisitor())


def scan_outputs(file: os.PathLike) -> list[str]:
    # Traverse the python file and find
    # all calls to `processor()`
    # function decorator

    return _scan_tree(file, OutputSearchVisitor())


def scan_inits(file: os.PathLike) -> list[str]:
    # Traverse the python file and find
    # all calls to `processor()`
    # function decorator

    return _scan_tree(file, InitSearchVisitor())


def scan(file: os.PathLike) -> list[str]:
    # Traverse the python file and find
    # all calls to `processor()`
    # function decorator

    return (
        scan_processors(file),
        scan_inputs(file),
        scan_outputs(file),
        scan_inits(file),
    )



import uuid
from collections import defaultdict
from typing import Any

import jls_utils as jls

from malevich._autoflow.tracer import tracer
from malevich._utility.registry import Registry
from malevich.constants import DEFAULT_CORE_HOST
from malevich.interpreter.abstract import Interpreter
from malevich.manifest import ManifestManager


class CoreInterpreterState:
    """State of the CoreInterpreter"""

    def __init__(self) -> None:
        # Involved operations
        self.ops: dict[str, Any] = {}
        # Dependencies (same keys as in self.ops)
        self.depends = defaultdict(list[tuple[tracer, str]])
        # Registry reference (just a shortcut)
        self.reg = Registry()
        # Manifest manager reference (just a shortcut)
        self.manf = ManifestManager()
        # Task configuration
        self.cfg = jls.Cfg()
        # Collections (key: operation_id, value: (`Collection` object, core_id,))
        self.collections: dict[str, str] = {}
        # Uploaded operations (key: operation_id, value: core_op_id)
        self.core_ops: dict[str, str] = {}
        # Interpreter parameters
        self.params: dict[str, Any] = {}


class CoreInterpreter(Interpreter[CoreInterpreterState, tuple[str, str]]):
    """Inteprets the flow using Malevich Core API"""

    def __init__(
        self, core_host: str = DEFAULT_CORE_HOST, core_auth: tuple[str, str] = None
    ) -> None:
        super().__init__(CoreInterpreterState())
        self.__core_host = core_host
        self.__core_auth = core_auth
        self._state.params["core_host"] = core_host
        self._state.params["core_auth"] = core_auth

        self.update_state()


    def create_node(
        self, state: CoreInterpreterState, tracer: tracer[tuple[str, dict]]
    ) -> CoreInterpreterState:
        state.ops[tracer.owner[0]] = {
            **state.reg.get(tracer.owner[0], {}),
            **tracer.owner[1],
            "tracer": tracer,
        }
        return state

    def create_dependency(
        self,
        state: CoreInterpreterState,
        callee: tracer[tuple[str, dict]],
        caller: tracer[tuple[str, dict]],
        link: Any,  # noqa: ANN401
    ) -> CoreInterpreterState:
        state.depends[caller.owner[0]].append((callee.owner[0], link))
        return state

    def before_interpret(self, state: CoreInterpreterState) -> CoreInterpreterState:
        jls.set_host_port(self.__core_host)
        if self.__core_auth is None:
            self.__core_auth = (uuid.uuid4().hex, uuid.uuid4().hex)
            jls.create_user(self.__core_auth)
        jls.update_core_credentials(self.__core_auth[0], self.__core_auth[1])
        state.params["core_auth"] = self.__core_auth
        return state

    def after_interpret(self, state: CoreInterpreterState) -> CoreInterpreterState:
        for id, app in state.ops.items():
            if operation_id := app.get("operation_id", None):
                __app = state.reg.get(operation_id)
                image_auth_user = state.manf.query(
                    *__app["image_auth_user"], resolve_secrets=True
                )

                image_auth_pass = state.manf.query(
                    *__app["image_auth_pass"], resolve_secrets=True
                )

                image_ref = state.manf.query(
                    *__app["image_ref"], resolve_secrets=True
                )

                extra_colls = {}

                for dependency, (link, _) in state.depends[id]:
                    if dependency in state.collections:
                        coll, __id = state.collections[dependency]
                        state.cfg.collections = {
                            **state.cfg.collections,
                            f"{coll.collection_id}": __id,
                        }
                        extra_colls[link] = coll.collection_id

                __app_id = uuid.uuid4().hex + f"-{__app['processor_id']}"
                jls.create_app(
                    app_id=__app_id,
                    processor_id=__app["processor_id"],
                    image_auth=(image_auth_user, image_auth_pass),
                    image_ref=image_ref,
                    extra_collections_from=extra_colls,
                    app_cfg=app.get("app_cfg", {}),
                )

                state.core_ops[id] = __app_id
            elif collection_ref := app.get("__collection__", None):
                if isinstance(app["tracer"], tracer):
                    __id = jls.create_collection_from_df(
                        collection_ref.collection_data,
                        collection_ref.collection_id,
                    )
                    state.collections[id] = (collection_ref, __id)
                else:
                    raise ValueError("Collection reference is not a tracer")
        return state

    def get_result(self, state: CoreInterpreterState) -> str:
        for id, core_id in sorted(
            state.core_ops.items(), key=lambda x: len(state.depends[x[0]])
        ):
            depends = state.depends[id]
            depends.sort(key=lambda x: x[1][1])
            depends = [
                state.core_ops[x[0]]
                for x in depends
                if x[0] in state.core_ops
            ]

            jls.create_task(
                task_id=core_id,
                app_id=core_id,
                tasks_depends=depends
            )

        __cfg = uuid.uuid4().hex
        jls.create_cfg(__cfg, state.cfg)
        leaves = [*self._tree.leaves()]
        return state.core_ops[leaves[0].owner[0]], __cfg



"""Installer interface."""

from abc import ABC, abstractmethod

from ..models.manifest import Dependency


class Installer(ABC):
    """Abstract installer interface."""
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def install(self, *args, **kwargs) -> Dependency: # noqa: ANN003, ANN002, ANN204
        """Installation function. Called by CLI commands"""
        pass

    @abstractmethod
    def restore(self, dependency: Dependency) -> None:
        """Restore function. Called by CLI commands with the Dependency object"""
        pass

    @abstractmethod
    def construct_dependency(self, object: dict) -> Dependency:
        """Construct a Dependency object from a dictionary"""
        pass



from typing import Any, Optional

from pydantic import BaseModel


class Dependency(BaseModel):
    package_id: str
    version: str
    installer: str
    options: dict[str, Any]


class Manifest(BaseModel):
    project_id : Optional[str] = None
    version: Optional[str] = None
    dependencies: list[dict[str, Dependency]] = []


class ManifestUpdateEntry(BaseModel):
    package_id: str
    version: str
    installer: str
    options: dict[str, Any]


class Secret(BaseModel):
    secret_key: str
    secret_value: str
    salt: Optional[str] = None
    external: Optional[bool] = False

class Secrets(BaseModel):
    secrets: dict[str, Secret] = {}






import os
import re
from copy import deepcopy
from typing import Any, Iterable

import pydantic_yaml as pydml

from malevich._utility.singleton import SingletonMeta
from malevich.models.manifest import Manifest, Secret, Secrets


class ManifestManager(metaclass=SingletonMeta):
    @staticmethod
    def secret_pattern() -> str:
        return r"secret#[0-9]{1,6}"

    @staticmethod
    def is_secret(value: str) -> bool:
        return re.match(ManifestManager.secret_pattern(), value) is not None

    def __init__(self) -> None:
        self.__path = os.path.join(os.getcwd(), "malevich.yaml")
        self.__secrets_path = os.path.join(os.getcwd(), "malevich.secrets.yaml")
        if not os.path.exists(self.__path):
            with open(self.__path, "w") as _file:
                pydml.to_yaml_file(_file, Manifest())
        if not os.path.exists(self.__secrets_path):
            with open(self.__secrets_path, "w") as _file:
                pydml.to_yaml_file(_file, Secrets())
        with open(self.__path) as _file:
            self.__manifest = pydml.parse_yaml_file_as(Manifest, _file)
            self.__backup = self.__manifest.model_dump()
            self.__secrets = pydml.parse_yaml_file_as(Secrets, self.__secrets_path)

    def cleanup_secrets(self) -> list[Secret]:
        secrets = re.findall(
            r"secret#[0-9]{1,6}", self.__manifest.model_dump_json(indent=4)
        )
        cleaned_secrets = Secrets(
            secrets={
                key: value
                for key, value in self.__secrets.model_dump()["secrets"].items()
                if key in secrets
            }
        )

        return cleaned_secrets

    def save(self) -> None:
        try:
            pydml.to_yaml_file(self.__path, self.__manifest)
            self.__backup = self.__manifest.model_dump()
        except Exception as e:
            pydml.to_yaml_file(self.__path, Manifest(**self.__backup))
            raise e
        finally:
            self.__manifest = pydml.parse_yaml_file_as(Manifest, self.__path)

        # self.__secrets = self.cleanup_secrets()
        pydml.to_yaml_file(self.__secrets_path, self.__secrets)
        self.__secrets = pydml.parse_yaml_file_as(Secrets, self.__secrets_path)

    def __query_list(
            self, __list: list[str | dict[str, Any]], __key: str
        ) -> Any:  # noqa: ANN401
        # Iterate over list object
        # within the Manifest
        # If list item is a dict, we need to check
        # if the key is in the dict

        # If there are multiple dicts with the same key,
        # we raise an error
        found = False
        for value in filter(lambda x: isinstance(x, dict), __list):
            if __key in value:
                if found:
                    raise ValueError(
                        f"Multiple values found for key {__key} in Manifest"
                    )
                found = True

        # If there are multiple strings with the same value,
        # we raise an error
        for value in filter(lambda x: isinstance(x, str), __list):
            if value == __key:
                if found:
                    raise ValueError(
                        f"Multiple values found for key {__key} in Manifest"
                    )
                found = True

        # After checking for multiple values,
        # we can iterate over the list again
        # and return the first value that matches
        for value in __list:
            if isinstance(value, dict) and __key in value:
                return value[__key]
            elif isinstance(value, str) and value == __key:
                return value
            elif isinstance(value, list):
                raise ValueError("Nested lists are not supported in Manifest")
        return None

    def query(
        self, *query: Iterable[str], resolve_secrets: bool = False
    ) -> Any:  # noqa: ANN401
        # Only iterate over dict representation of Manifest
        # for simplicity
        cursor = self.__manifest.model_dump()
        for key in query:
            # If the cursor is a list, we need to iterate over it
            if isinstance(cursor, list):
                cursor = self.__query_list(cursor, key)
            elif isinstance(cursor, dict):
                cursor = cursor[key]
            else:
                return None
        if (
            cursor is not None
            and resolve_secrets
            and isinstance(cursor, str)
            and re.match(r"^secret#[0-9]{1,6}", cursor)
        ):
            return self.query_secret(cursor).secret_value
        return cursor

    def put(
        self, *path: Iterable[str], value: Any, append: bool = False  # noqa: ANN401
    ) -> Manifest:  # noqa: ANN401
        pre = path[:-1]
        key = path[-1]
        dump = self.__manifest.model_dump()
        cursor = dump
        for q in pre:
            if isinstance(cursor, list):
                try:
                    q = int(q)
                    __cursor = cursor[q]
                except ValueError:
                    pass
                else:
                    cursor = __cursor
                    continue

                for i, _c in enumerate(cursor):
                    if q in _c:
                        cursor = cursor[i][q]
                        break
                else:
                    cursor.append({})
                    cursor = cursor[-1]
                    continue
            if q not in cursor:
                cursor[q] = {}
            cursor = cursor[q]
        if isinstance(cursor, list):
            if append:
                cursor.append(value)
            else:
                for i, _c in enumerate(cursor):
                    if key in _c:
                        cursor[i][key] = value
                        break
        else:
            if append:
                cursor[key].append(value)
            else:
                cursor[key] = value
        self.__manifest = Manifest(**dump)
        self.save()
        return self.__manifest

    def put_secret(
        self,
        key: str,
        value: str,
        salt: str = None,
        external: bool = False
    ) -> str:
        # Create new secret and hash it with 6 digits
        if salt:
            secret_hash_id = hash(f'{key}="{value}"{salt}'.encode()) % 1000000
        else:
            secret_hash_id = hash(f'{key}="{value}"'.encode()) % 1000000
        __key = f"secret#{secret_hash_id}"
        secrets = self.__secrets.model_dump()
        secrets["secrets"][__key] = Secret(
            secret_key=key, secret_value=value, salt=salt, external=external
        )
        self.__secrets = Secrets(**secrets)
        return __key

    def put_secrets(self, salt: str = None, **kwargs: dict[str, str]) -> list[str]:
        secrets = []
        for key, value in kwargs.items():
            if value:
                secrets.append(self.put_secret(key, value, salt))
            else:
                secrets.append(None)
        return secrets

    def query_secret(self, key: str, only_value: bool = False) -> Secret | None:
        _s =  self.__secrets.secrets.get(key, None)
        if _s and only_value:
            return _s.secret_value
        return _s

    def as_dict(self) -> dict[str,]:
        return deepcopy(self.__manifest.model_dump())

    def get_secrets(self) -> Secrets:
        return deepcopy(self.__secrets.model_dump())




from typing import Callable, List, Optional


def input_doc(id: Optional[str] = None, collection_from: Optional[str] = None, collections_from: Optional[List[str]] = None, extra_collection_from: Optional[str] = None, extra_collections_from: Optional[List[str]] = None, query: Optional[str] = None, finish_msg: Optional[str] = None):  # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def input_df(id: Optional[str] = None, collection_from: Optional[str] = None, collections_from: Optional[List[str]] = None, extra_collection_from: Optional[str] = None, extra_collections_from: Optional[List[str]] = None, by_args: bool = False, query: Optional[str] = None, finish_msg: Optional[str] = None): # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def input_true(id: Optional[str] = None, collection_from: Optional[str] = None, collections_from: Optional[List[str]] = None, extra_collection_from: Optional[str] = None, extra_collections_from: Optional[List[str]] = None, query: Optional[str] = None, finish_msg: Optional[str] = None):  # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def processor(id: Optional[str] = None, finish_msg: Optional[str] = None, drop_internal: bool = True, get_scale_part_all: bool = False):    # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def output(id: Optional[str] = None, collection_name: Optional[str] = None, collection_names: Optional[List[str]] = None, finish_msg: Optional[str] = None, drop_internal: bool = True):    # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper


def scheme():  # noqa: ANN201
    def wrapper(cl):  # noqa: ANN202, ANN001
        return cl
    return wrapper


def init(id: Optional[str] = None, enable: bool = True, tl: Optional[int] = None, prepare: bool = False):   # noqa: ANN201, E501
    def wrapper(fun: Callable) -> Callable:
        return fun
    return wrapper
